#include "Application.h"
#include "MainWindow.h"
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include "TPS.h"
#include "ShapeContext.h"
#include "Munkres.h"
#include "CannyEdgeDetector.h"
#include "Image.h"
#include "Utility.h"


/*****************************************************************************/
/* Inicialização da aplicação
/* Responsável por exibir a janela principal
/*****************************************************************************/
bool Application::OnInit() {
    wxInitAllImageHandlers();
    mainWindow = new MainWindow(0, 0, wxPoint(100,100),wxSize(500,300));
    mainWindow->Show(true);
    mainWindow->PointsBoxSizer->RecalcSizes();
    CannyEdgeDetector *canny = CannyEdgeDetector::getInstance();
    canny->setParameters(1.5,180,200);
    img1 = NULL;
    return true;
}

/*****************************************************************************/
/* Devolve a imagem
/*****************************************************************************/
Image* Application::getImage() {
    return img1;
}

/*****************************************************************************/
/* Devolve a imagem
/*****************************************************************************/
Image* Application::getImage2() {
    return img2;
}

/*****************************************************************************/
/* Devolve o conjunro de imagens
/*****************************************************************************/
std::vector<Image*> Application::getImages() {
    return imgs;
}

/*****************************************************************************/
/* Define o conjunto de imagens
/*****************************************************************************/
void Application::setImages(std::vector<Image*> imgs) {
    this->imgs = imgs;
}

/*****************************************************************************/
/* Define a imagem
/*****************************************************************************/
void Application::setImage1(Image* img) {
    img1 = img;

    wxBitmap b1 = wxBitmap(img->getOriginal());
    mainWindow->bitmapOriginal->SetBitmap(b1);

    mainWindow->bitmapContour1->SetBitmap(img->getContourBitmap());
    mainWindow->bitmapPoints1->SetBitmap(img->getContourPointsBitmap(SAMPLE_SIZE));
}

/*****************************************************************************/
/* Define a imagem
/*****************************************************************************/
void Application::setImage2(Image* img) {
    img2 = img;
}

/*****************************************************************************/
/* Executa
/*****************************************************************************/
void Application::run() {
    int i, j;
    std::clock_t begin = clock();
    std::vector<Point> pts1, pts2;

    int n1 = SAMPLE_SIZE, n2 = SAMPLE_SIZE;

    if (n1 > (int)(img1->getAllPoints().size())) n1 = (int)(img1->getAllPoints().size());
    if (n2 > (int)(img2->getAllPoints().size())) n2 = (int)(img2->getAllPoints().size());


    img1->computeShapeContexts(n1, median(img1->getAllPoints()));
    img2->computeShapeContexts(n2, median(img2->getAllPoints()));
    std::vector<ShapeContext*> SC1 = img1->getShapeContexts();
    std::vector<ShapeContext*> SC2 = img2->getShapeContexts();

    for (unsigned int i = 0; i < SC1.size(); i++)
        pts1.push_back(SC1[i]->getPoint());

    for (unsigned int i = 0; i < SC2.size(); i++)
        pts2.push_back(SC2[i]->getPoint());


    matrix<double> Cost (SC1.size(),SC2.size());
    for (unsigned int i = 0; i < SC1.size(); i++) {
        ShapeContext *sc1 = SC1[i];
        for (unsigned int j = 0; j < SC2.size(); j++) {
            ShapeContext *sc2 = SC2[j];
            Cost(i,j) = sc1->distance(sc2);
        }
    }

    vector<int> results(Cost.size1());

    munkresSolver.compute(Cost,results);

    optimalCost = 0;
    for (unsigned i = 0; i < results.size(); i++)
        optimalCost += Cost(i,results(i));


    mainWindow->lblOptimalCost->SetLabel(wxString::Format(wxT("Optimal Cost: %f"),optimalCost));


    if (imgs.size() == 1) {
        wxBitmap b2 = wxBitmap(imgs[0]->getOriginal());
        mainWindow->bitmapOriginal2->SetBitmap(b2);
        mainWindow->bitmapContour2->SetBitmap(imgs[0]->getContourBitmap());
        mainWindow->bitmapPoints2->SetBitmap(imgs[0]->getContourPointsBitmap(SAMPLE_SIZE));
        mainWindow->PointsBoxSizer->Fit(mainWindow->PointsPanel);
        // Tabela de custos
        wxFont font(10, wxFONTFAMILY_DEFAULT, wxNORMAL, wxNORMAL, false, wxT("Arial"), wxFONTENCODING_ISO8859_1);
        wxFont fontBold(10, wxFONTFAMILY_DEFAULT, wxNORMAL, wxBOLD, false, wxT("Arial"), wxFONTENCODING_ISO8859_1);
        for (unsigned i = 0; i < Cost.size1(); i++) {
            for (unsigned j = 0; j < Cost.size2(); j++) {
                mainWindow->costGrid->SetCellFont(i,j,font);
            }
        }
        for (unsigned i = 0; i < Cost.size1(); i++) {
            mainWindow->costGrid->SetColLabelValue(i, wxString::Format(wxT("(%d,%d)"), pts1[i].x, pts1[i].y));
            for (unsigned j = 0; j < Cost.size2(); j++) {
                mainWindow->costGrid->SetRowLabelValue(j, wxString::Format(wxT("(%d,%d)"), pts2[j].x, pts2[j].y));
                mainWindow->costGrid->SetCellValue(i,j,wxString::Format(wxT("%.2f"),Cost(i,j)));
                mainWindow->costGrid->SetCellFont(i,results(i),fontBold);
            }
        }

        // Construcao da imagem com a correspondencia de pontos
        int w = img1->getWidth() + img2->getWidth() + 200;
        int h = img1->getHeight() + img2->getHeight() + 200;
        wxBitmap bmp = wxBitmap(w,h);
        wxMemoryDC dc;
        dc.SelectObject(bmp);

        dc.Clear();
        for (unsigned i = 0; i < SAMPLE_SIZE; i++) {
            Point pt1 = pts1[i];
            Point pt2 = pts2[results(i)];
            dc.SetBrush(wxBrush(*wxBLUE, wxSOLID));
            dc.DrawCircle(pt1.x, pt1.y, 2);
            dc.SetBrush(wxBrush(*wxGREEN, wxSOLID));
            dc.DrawCircle(pt2.x+200, pt2.y+20, 2);
            dc.DrawLine(pt1.x, pt1.y, pt2.x+200, pt2.y+20);
        }
        mainWindow->correspondenceBitmap->SetBitmap(bmp);
    }
}


long double Application::getBendingEnergy() {
    return bendingEnergy;
}

double Application::getOptimalCost() {
    return optimalCost;
}

TPS Application::getTPS() {
    return tps;
}

Application::Application() {

}
