#include <stdlib.h>
#include <stdio.h>
#include "wx/wx.h"
#include "CannyEdgeDetector.h"


//*****************************************************************************/
//* Retorna o valor da função de gauss
//*****************************************************************************/
float CannyEdgeDetector::gauss(float x, float sigma) {
    if (sigma == 0) return 0.0;
    return (float) exp ((double)((-x*x)/(2*sigma*sigma)));
}

//*****************************************************************************/
//* Retorna a primeira derivada da função de gauss
//*****************************************************************************/
float CannyEdgeDetector::dGauss (float x, float sigma) {
    return -x/(sigma*sigma)*gauss(x, sigma);
}

//*****************************************************************************/
//* Aplica a convolução da imagem im com uma máscara de tamanho
//* width contendo a função de gauss nas direções X e Y
//*****************************************************************************/
void CannyEdgeDetector::seperable_convolution (shared_ptr<wxImage>&img, float *mask, int width, matrix<float>&smx, matrix<float>&smy) {
    int i, j, k, I1, I2;
    float x, y;

    int rows = img->GetHeight();
    int cols = img->GetWidth();

    unsigned char* data = img->GetData();

    for (i = 0; i < rows; i++)
        for (j = 0; j < cols; j++) {
            x = mask[0] * data[3*(i*cols+j)+1];
            y = mask[0] * data[3*(i*cols+j)+1];
            for (k = 1; k < width; k++) {
                I1 = (i+k) % rows;
                I2 = (i-k+rows) % rows;
                y += mask[k] * data[3*(I1*cols+j)+1] + mask[k] * data[3*(I2*cols+j)+1];
                I1 = (j+k) % cols;
                I2 = (j-k+cols) % cols;
                x += mask[k] * data[3*(i*cols+I1)+1] + mask[k] * data[3*(i*cols+I2)+1];
            }
            smx(i,j) = x;
            smy(i,j) = y;
        }
}

//*****************************************************************************/
//* Aplica a convolução da imagem im com uma máscara de tamanho
//* width contendo a primeira derivada da função de gauss
//* nas direções X e Y
//*****************************************************************************/
void CannyEdgeDetector::dxy_seperable_convolution
(matrix<float>&imx, matrix<float> &imy, int nr, int nc, float *mask, int width, matrix<float>&smx, matrix<float>&smy) {
  int i, j, k, I1, I2;
  float x, y;

  for (i = 0; i < nr; i++)
    for (j = 0; j < nc; j++) {
      x = 0.0;
      y = 0.0;
      for (k = 1; k < width; k++) {
        I1 = (i+k) % nr;
        I2 = (i-k+nr) % nr;
        y += -mask[k] * imy(I1,j) + mask[k] * imy(I2,j);
        I1 = (j+k) % nc;
        I2 = (j-k+nc) % nc;
        x += -mask[k] * imx(i,I1) + mask[k] * imx(i,I2);
      }
      smx(i,j) = x;
      smy(i,j) = y;
    }
}

//*****************************************************************************/
//* Devolve a norma de um vetor com coordenadas x e y
//*****************************************************************************/
float CannyEdgeDetector::norm (float x, float y) {
    return (float) sqrt ((double)(x*x + y*y));
}

//*****************************************************************************/
//* nonmax_supress: Realiza a supressão não maximal
//*****************************************************************************/
void CannyEdgeDetector::nonmax_suppress (shared_ptr<wxImage> &mag_img, matrix<float>&dx, matrix<float>&dy, int nr, int nc) {
    int i, j;
    float xx, yy, g2, g1, g3, g4, g, xc, yc;

    unsigned char* mag_data = mag_img->GetData();

    int rows = mag_img->GetHeight();
    int cols = mag_img->GetWidth();

    for (i = 1; i < rows-1; i++) {
        for (j = 1; j < cols-1; j++) {
            /* Trata as derivadas x e y como componentes de um vetor */
            xc = dx(i,j);
            yc = dy(i,j);
            g  = norm (xc, yc);

            if (fabs (yc) > fabs (xc)) {
                /* A compontente Y é a maior, então a direção do gradiente é basicamente CIMA/BAIXO */
                xx = fabs (xc) / fabs (yc);
                yy = 1.0;

                g2 = norm (dx(i-1,j), dy(i-1,j));
                g4 = norm (dx(i+1,j), dy(i+1,j));
                if (xc*yc > 0.0) {
                    g3 = norm (dx(i+1,j+1), dy(i+1,j+1));
                    g1 = norm (dx(i-1,j-1), dy(i-1,j-1));
                } else {
                    g3 = norm (dx(i+1,j-1), dy(i+1,j-1));
                    g1 = norm (dx(i-1,j+1), dy(i-1,j+1));
                }

              } else {
                  /* A compontente X é a maior, então a direção do gradiente é basicamente ESQ/DIR */
                  xx = fabs (yc) / fabs (xc);
                  yy = 1.0;

                  g2 = norm (dx(i,j+1), dy(i,j+1));
                  g4 = norm (dx(i,j-1), dy(i,j-1));
                  if (xc * yc > 0.0) {
                    g3 = norm (dx(i-1,j-1), dy(i-1,j-1));
                    g1 = norm (dx(i+1,j+1), dy(i+1,j+1));
                }
                else {
                    g1 = norm (dx(i-1,j+1), dy(i-1,j+1));
                    g3 = norm (dx(i+1,j-1), dy(i+1,j-1));
                }
            }

            /* Agora determina se o pixel atual é máximo */
        unsigned char value;
   	    if ((g > (xx * g1 + (yy - xx) * g2)) && (g > (xx * g3 + (yy - xx) * g4))) {
                if (g <= 255)
                   value = (unsigned char) g;
                else
                   value = 255;
        } else {
            value = 0;
        }
        mag_data[3*(i*cols+j)] = value;
        mag_data[3*(i*cols+j)+1] = value;
        mag_data[3*(i*cols+j)+2] = value;
    }

    }
}

//*****************************************************************************/
//* Verifica se está no intervalo
//*****************************************************************************/
bool CannyEdgeDetector::range (shared_ptr<wxImage>img, int i, int j) {
    if ((i < 0) || (i >= img->GetHeight())) return false;
    if ((j < 0) || (j >= img->GetWidth())) return false;
    return true;
}

//*****************************************************************************/
//* Função auxiliar para o passo da histerese:
//*****************************************************************************/
int CannyEdgeDetector::trace (int i, int j, unsigned char low, shared_ptr<wxImage>&img, shared_ptr<wxImage>&mag) {
    int n, m;
    char flag = 0;

    //int rows = img->GetHeight();
    int cols = img->GetWidth();

    unsigned char *data = img->GetData();
    unsigned char *mag_data = mag->GetData();

    if (data[3*(i*cols+j)+1] == 255) {
        data[3*(i*cols+j)] = 0;
        data[3*(i*cols+j)+1] = 0;
        data[3*(i*cols+j)+2] = 0;
        flag = 0;
        for (n = -1; n <= 1; n++) {
            for (m = -1; m <= 1; m++) {
                if (i == 0 && m == 0) continue;
                if (range(mag, i+n, j+m) && mag_data[3*((i+n)*cols+(j+m)) + 1] >= low)
                    if (trace(i+n, j+m, low, img, mag)) {
                        flag = 1;
                        break;
                    }
            }
            if (flag) break;
        }
        return(1);
    }
    return(0);
}

//*****************************************************************************/
//* Realiza o que Canny chamou de histerese na imagem.
//*   1) Marca todas as bordas com magnitude maior que high como corretas.
//*   2) Percorre todos os pixels com magnitude de borda E[low, high].
//*   3) Se um pixel está conectado a outro já marcado como borda,
//*      marca-o também.
//*****************************************************************************/
void CannyEdgeDetector::hysteresis (shared_ptr<wxImage>&mag, unsigned char high, unsigned char low, shared_ptr<wxImage>&img) {
    int i, j;

    int rows = img->GetHeight();
    int cols = img->GetWidth();

    unsigned char *data = img->GetData();
    unsigned char *mag_data = mag->GetData();

    /* Limpa a matriz da imagem */
    for (i = 0; i < rows; i++)
         for (j = 0; j < cols; j++) {
              data[3*(i*cols+j)] = 255;
              data[3*(i*cols+j)+1] = 255;
              data[3*(i*cols+j)+2] = 255;
         }

    /* Traça arestas ao longo de todos os caminhos cujos valores da magnitude
       mantem se acima do valor de menor threshold */
    for (i = 0; i < rows; i++)
        for (j = 0; j < cols; j++)
            if (mag_data[3*(i*cols+j)+1] >= high)
                trace (i, j, low, img, mag);
}

//*****************************************************************************/
//* Aplica o detector de bordas de Canny na imagem img, onde s é o sigma
//* da função de Gauss, low e high são os menores e maiores valores para
//* threshold respectivamente.
//*****************************************************************************/
void CannyEdgeDetector::run (shared_ptr<wxImage>&img) {
    int i, j, width = 0;
    float gau[20], dgau[20];

    unsigned char* data = img->GetData();
    int rows = img->GetHeight();
    int cols = img->GetWidth();

    shared_ptr<wxImage> mag_img(new wxImage(cols,rows, true));
    unsigned char* mag_data = mag_img->GetData();

    /* Cria o filtro da gaussiana e a derivada do filtro da gaussiana */
    for (i = 0; i < 20; i++) {
        gau[i] = gauss ((float)i, sigma);
        if (gau[i] < 0.0005) {
            width = i;
            break;
        }
        dgau[i] = dGauss ((float)i, sigma);
    }
    /* Convolução da imagem com a gaussiana nas direções X e Y */
    matrix<float> smx(rows,cols);
    matrix<float> smy(rows,cols);

    seperable_convolution (img, gau, width, smx, smy);

    /* Convolução da imagem com ruido com a derivada na direção X e Y */
    matrix<float> dx(rows,cols);
    matrix<float> dy(rows,cols);
    dxy_seperable_convolution (smx, smy, rows, cols, dgau, width, dx, dy);

    for (i = 0; i < rows; i++)
        for (j = 0; j < cols; j++) {
            data[3*(i*cols+j)] = mag_data[3*(i*cols+j)];
            data[3*(i*cols+j)+1] = mag_data[3*(i*cols+j)+1];
            data[3*(i*cols+j)+2] = mag_data[3*(i*cols+j)+2];
        }

    /* Supressão não maximal - pixels da aresta devem ser um máximo local */
    nonmax_suppress (mag_img, dx, dy, rows, cols);
    /* Histerese dos pixels da aresta */
    hysteresis (mag_img, high, low, img);

}

