#include "Munkres.h"
#include <iostream>

//*****************************************************************************/
// Implementação do algoritmo Munkres baseado em:
// http://csclab.murraystate.edu/bob.pilgrim/445/munkres.html
//*****************************************************************************/

using namespace boost::numeric::ublas;

void Munkres::step1 () {
	double minValue;
	unsigned int i, j;

	for (i = 0; i < Cost.size1(); i++) {
		minValue = Cost(i,0);
		for (j = 1; j < Cost.size2(); j++) {
			if (Cost(i,j) < minValue)
				minValue = Cost(i, j);
		}
		for (j = 0; j < Cost.size2(); j++) {
			Cost(i,j) = Cost(i,j) - minValue;
		}
	}
	step = 2;
}

void Munkres::step2 () {
    unsigned int i, j;
	for (i = 0; i < Cost.size1(); i++) {
		colCov[i] = 0;
		rowCov[i] = 0;
	}
	for (i = 0; i < Cost.size1(); i++) {
		for (j = 0; j < Cost.size2(); j++) {
		    M(i,j) = 0;
			if ((Cost(i,j) == 0) &&  (colCov[j] == 0) && (rowCov[i] == 0)) {
				M(i,j) = 1;
				colCov[j] = 1;
				rowCov[i] = 1;
			}
		}
	}

	for (i = 0; i < Cost.size1(); i++) {
		colCov[i] = 0;
		rowCov[i] = 0;
	}
	step = 3;
}


void Munkres::step3 () {
	for (unsigned i = 0; i < Cost.size1(); i++)
		for (unsigned j = 0; j < Cost.size2(); j++)
			if (M(i,j) == 1)
				colCov[j] = 1;

	unsigned count = 0;
	for (unsigned j = 0; j < Cost.size2(); j++)
		count = count + colCov[j];

	if (count == Cost.size2())
		step = 7;
	else
		step = 4;
}


void Munkres::find_a_zero(int &row, int &col) {
    int i, j;
	row = -1;
	col = -1;
	int m = (int)Cost.size1();
	int n = (int)Cost.size2();
	for (i = 0; i < m; ++i) {
	    if (rowCov[i] == 0) {
            for (j = 0; j < n; ++j) {
                if ((Cost(i,j) == 0) && (colCov[j] == 0)) {
                    row = i;
                    col = j;
                    break;
                }
            }
	    }
	}
}

int Munkres::star_in_row(int row) {
	unsigned int j;
	int tbool = 0;
	for (j = 0; j < Cost.size2(); j++) {
		if (M(row,j) == 1)
			tbool = 1;
	}
	return tbool;
}

void Munkres::find_star_in_row (int row, int *col) {
	unsigned int j;
	*col = -1;
	for (j = 0; j < Cost.size2(); j++) {
		if (M(row,j) == 1)
			*col = j;
	}
}

void Munkres::step4 () {
	int i, j, row, col;
	int done;
	done = false;

	int m = (int)Cost.size1();
	int n = (int)Cost.size2();

	while (!done) {
        row = -1;
        col = -1;

        for (i = 0; i < m; ++i) {
            if (rowCov[i] == 0) {
                for (j = 0; j < n; ++j) {
                    if ((Cost(i,j) == 0) && (colCov[j] == 0)) {
                        row = i;
                        col = j;
                        break;
                    }
                }
            }
        }

		if (row == -1) {
			done = true;
			step = 6;
		}
		else {
			M(row,col) = 2;
			if (star_in_row(row)) {
				find_star_in_row(row,&col);
				rowCov[row] = 1;
				colCov[col] = 0;
			}
			else {
				done = true;
				step = 5;
				Z0_r = row;
				Z0_c = col;
			}
		}
	}
}

void Munkres::find_star_in_col(int c, int *r) {
	unsigned int i;
	*r = -1;
	for (i = 0; i < Cost.size1(); i++)
      if (M(i,c) == 1)
        *r = i;
}

void Munkres::find_prime_in_row(int r, int *c) {
	unsigned int j;
	for (j = 0; j < Cost.size2(); j++)
      if (M(r,j) == 2)
        *c = j;
}

void Munkres::convert_path(int count) {
	int i;
	for (i = 0; i <= count; i++) {
		if (M(path(i,0),path(i,1)) == 1)
        M(path(i,0),path(i,1)) = 0;
      else
        M(path(i,0),path(i,1)) = 1;
	}
}

void Munkres::clear_covers() {
	unsigned int i;
	for (i = 0; i < Cost.size1(); i++) {
		rowCov[i] = 0;
      colCov[i] = 0;
	}
}

void Munkres::erase_primes() {
	for (unsigned i = 0; i < Cost.size1(); i++)
		for (unsigned j = 0; j < Cost.size2(); j++)
			if (M(i,j) == 2)
				M(i,j) = 0;
}

void Munkres::step5() {
	int count;
	int r,c = 0;
	int done;
	count = 0;
	path(count,0) = Z0_r;
	path(count,1) = Z0_c;
	done = false;
	while (!done) {
		find_star_in_col(path(count,1), &r);
		if (r > -1) {
			count = count + 1;
			path(count,0) = r;
			path(count,1) = path(count-1,1);
		}
		else
			done = true;
		if (!done) {
			find_prime_in_row(path(count,0),&c);
			count = count + 1;
			path(count,0) = path(count-1,0);
			path(count,1) = c;
		}
	}

	convert_path(count);
	clear_covers();
	erase_primes();
	step = 3;
}

double Munkres::find_smallest() {
	long double minVal = INT_MAX;
	for (unsigned i = 0; i < Cost.size1(); ++i)
        if (rowCov[i] == 0) {
            for (unsigned j = 0; j < Cost.size2(); ++j)
                if (colCov[j] == 0)
                    if (Cost(i,j) < minVal)
                        minVal = Cost(i,j);
        }
    return minVal;
}

void Munkres::step6 () {
	long double minVal = find_smallest();
	for (unsigned i = 0; i < Cost.size1(); ++i) {
		for (unsigned j = 0; j < Cost.size2(); ++j) {
			if (rowCov[i] == 1)
				Cost(i,j) = Cost(i,j) + minVal;
			if (colCov[j] == 0)
				Cost(i,j) = Cost(i,j) - minVal;
		}
	}
	step = 4;
}

void Munkres::compute(const matrix<double> &Cost, vector<int> &results) {
	unsigned int i, j;
	step = 1;

	unsigned int dim;
	if (Cost.size1() >= Cost.size2()) {
	    dim = Cost.size1();
	}
	else
        dim = Cost.size2();

    matrix<double> SquareCost = matrix<double> (dim,dim);
    for (unsigned int i = 0; i < Cost.size1(); i++)
        for (unsigned int j = 0; j < Cost.size2(); j++)
            SquareCost(i,j) = Cost(i,j);


    if (Cost.size1() > Cost.size2()) {
        for (unsigned int i = Cost.size1() - 1; i < dim; i++)
            for (unsigned int j = 0; j < Cost.size2(); j++)
                SquareCost(i,j) = 0;
    }
    else if (Cost.size2() > Cost.size1()) {
        for (unsigned int i = 0; i < Cost.size1(); i++)
            for (unsigned int j = Cost.size2() - 1; j < dim; j++)
                SquareCost(i,j) = 0;
    }

    this->Cost = SquareCost;
    this->M = matrix<double> (this->Cost.size1(), this->Cost.size2());

    rowCov = std::vector<int> (this->Cost.size1());
    colCov = std::vector<int>(this->Cost.size2());

    path = matrix<int> (this->Cost.size1()*this->Cost.size2(),2);
	for (i = 0; i < this->Cost.size1(); ++i) {
		rowCov[i] = colCov[i] = 0;
	}

    bool done = false;
    while (!done) {
        switch (step) {
            case 1 : step1(); break;
            case 2 : step2(); break;
            case 3 : step3(); break;
            case 4 : step4(); break;
            case 5 : step5(); break;
            case 6 : step6(); break;
            default: done = true;
        }
    }

	for (i = 0; i < M.size1(); ++i)
		for (j = 0; j < M.size2(); ++j)
			if (M(i,j))
			    results[i] = j;

    rowCov.clear();
    colCov.clear();
    M.clear();
}

